"""
Originally written by Micah J. Amended for uavsar_pytools by Zach Keskinen.
Functions convert polsar, insar, and other associated UAVSAR files from binary format to geoTIFFS in WGS84.
"""

import os
from os.path import isdir, exists, basename, dirname, join, isfile
from glob import glob
from tqdm import tqdm
import numpy as np
import pandas as pd
import pytz
import rasterio
from rasterio.transform import Affine
from rasterio.crs import CRS
from pyproj import Geod, Proj
import logging

log = logging.getLogger(__name__)
logging.basicConfig()
log.setLevel(logging.WARNING)

def get_encapsulated(str_line, encapsulator):
    """
    Returns items found in the encapsulator, useful for finding units
    Args:
        str_line: String that has encapusulated info we want removed
        encapsulator: string of characters encapusulating info to be removed
    Returns:
        result: list of strings found inside anything between encapsulators
    e.g.
        line = 'density (kg/m^3), temperature (C)'
        ['kg/m^3', 'C'] = get_encapsulated(line, '()')
    """

    result = []

    if len(encapsulator) > 2:
        raise ValueError('encapsulator can only be 1 or 2 chars long!')

    elif len(encapsulator) == 2:
        lcap = encapsulator[0]
        rcap = encapsulator[1]

    else:
        lcap = rcap = encapsulator

    # Split on the lcap
    if lcap in str_line:
        for i, val in enumerate(str_line.split(lcap)):
            # The first one will always be before our encapsulated
            if i != 0:
                if lcap != rcap:
                    result.append(val[0:val.index(rcap)])
                else:
                    result.append(val)

    return result

def read_annotation(ann_file):
    """
    .ann files describe the INSAR data. Use this function to read all that
    information in and return it as a dictionary
    Expected format:
    `DEM Original Pixel spacing (arcsec) = 1`
    Where this is interpretted as:
    `key (units) = [value]`
    Then stored in the dictionary as:
    `data[key] = {'value':value, 'units':units}`
    values that are found to be numeric and have a decimal are converted to a
    float otherwise numeric data is cast as integers. Everything else is left
    as strings.
    Args:
        ann_file: path to UAVSAR annotation file
    Returns:
        data: Dictionary containing a dictionary for each entry with keys
              for value, units and comments
    """

    with open(ann_file) as fp:
        lines = fp.readlines()
        fp.close()
    data = {}

    # loop through the data and parse
    for line in lines:

        # Filter out all comments and remove any line returns
        info = line.strip().split(';')
        comment = info[-1].strip().lower()
        info = info[0]
        # ignore empty strings
        if info and "=" in info:
            d = info.split('=')
            name, value = d[0], d[1]
            # Clean up tabs, spaces and line returns
            key = name.split('(')[0].strip().lower()
            units = get_encapsulated(name, '()')
            if not units:
                units = None
            else:
                units = units[0]

            value = value.strip()

            # Cast the values that can be to numbers ###
            if value.strip('-').replace('.', '').isnumeric():
                if '.' in value:
                    value = float(value)
                else:
                    value = int(value)

            # Assign each entry as a dictionary with value and units
            data[key] = {'value': value, 'units': units, 'comment': comment}

    # Convert times to datetimes
    if 'start time of acquistion for pass 1' in data.keys():
        for pass_num in ['1', '2']:
            for timing in ['start', 'stop']:
                key = f'{timing} time of acquisition for pass {pass_num}'
                dt = pd.to_datetime(data[key]['value'])
                dt = dt.astimezone(pytz.timezone('US/Mountain'))
                data[key]['value'] = dt
    elif 'start time of acquisition' in data.keys():
        for timing in ['start', 'stop']:
                key = f'{timing} time of acquisition'
                dt = pd.to_datetime(data[key]['value'])
                dt = dt.astimezone(pytz.timezone('US/Mountain'))
                data[key]['value'] = dt

    return data

def grd_tiff_convert(in_fp, out_dir, ann_fp = None, overwrite = 'user'):
    """
    Converts a single binary image either polsar or insar to geotiff.
    See: https://uavsar.jpl.nasa.gov/science/documents/polsar-format.html for polsar
    and: https://uavsar.jpl.nasa.gov/science/documents/rpi-format.html for insar
    and: https://uavsar.jpl.nasa.gov/science/documents/stack-format.html for SLC stacks.

    Args:
        in_fp (string): path to input binary file
        out_dir (string): directory to save geotiff in
        ann_fp (string): path to UAVSAR annotation file
    """
    out_fp = join(out_dir, basename(in_fp)) + '.tiff'

    # Determine type of image
    if isfile(out_dir):
        raise Exception('Provide filepath not the directory.')

    if not exists(in_fp):
        raise Exception(f'Input file path: {in_fp} does not exist.')

    extens = basename(in_fp).split('.')[1:]
    if len(extens) == 1:
        type = extens[0]
        subtype = None
    elif len(extens) == 2:
        subtype = extens[0]
        type = extens[1]
    else:
        raise Exception('Can only handle one or two extensions on input file')

    # Find annotation file in same directory if no user given one
    if not ann_fp:
        if subtype:
            ann_fp = in_fp.replace(f'.{subtype}', '').replace(f'.{type}', '.ann')
        else:
            ann_fp = in_fp.replace(f'.{type}', '.ann')
        if not exists(ann_fp):
            search_base = '_'.join(basename(in_fp).split('.')[0].split('_')[:4])
            search_full = os.path.join(os.path.dirname(in_fp), f'{search_base}*.ann')
            log.debug(f'Glob search: {search_full}')
            ann_search = glob(search_full)
            if len(ann_search) == 1:
                ann_fp = ann_search[0]
            else:
                raise Exception('No ann file found in directory. Please specify ann filepath.')
        else:
            log.info(f'No annotation file path specificed. Using f{ann_fp}.')

    # Check for compatible extensions
    if type == 'zip':
        raise Exception('Can not convert zipped directories. Unzip first.')
    if type == 'dat' or type == 'kmz' or type == 'kml' or type == '.png' or type == 'tif':
        raise Exception(f'Can not handle {type} products')
    if type == 'slc' or type == 'mlc':
        raise Exception('Unable to convert slant range products to WGS84. Download and convert .grd file.')
    if subtype == 'kmz':
        raise Exception('Can not handle kmz interferograms.')
    if type == 'ann':
        raise Exception(f'Can not convert annotation files.')

    # Check for slant range files and ancillary files
    anc = None
    if type == 'slope' or type == 'hgt' or type == 'inc':
        anc = True

    # Check if file already exists and for overwriting
    ans = 'N'
    if exists(out_fp):
            if overwrite == True:
                ans = 'y'
            elif overwrite == False:
                ans = 'n'
            else:
                ans = input(f'\nWARNING! You are about overwrite {out_fp}!.  '
                            f'\nPress Y to continue and any other key to abort: ').lower()
            if ans == 'y':
                os.remove(out_fp)

    if ans == 'y' or exists(out_fp) == False:

        # Read in annotation file
        desc = read_annotation(ann_fp)
        #pd.DataFrame.from_dict(desc).to_csv('../data/test.csv')
        if 'flight id for pass 2' in desc.keys():
            mode = 'insar'
        else:
            mode = 'polsar'
        log.info(f'Working with {mode}')

        # Determine the correct file typing for searching our data dictionary
        slant = None
        if not anc:
            if mode == 'polsar':
                polarization = basename(in_fp).split('_')[5][-4:]
                if type == 'grd':
                    if polarization == 'HHHH' or polarization == 'HVHV' or polarization == 'VVVV':
                            type = f'{type}_pwr'
                    else:
                        type = f'{type}_mag'
                # type = f'{type}_pwr'

            elif mode == 'insar':
                if type == 'int':
                    type = 'grd_phs'
                else:
                    type = 'grd'
                if subtype == None:
                    raise Exception('Unable to convert slant range. Download and convert .grd file.')

        log.debug(f'type: {type}')

        # Pull the appropriate values from our annotation dictionary
        nrow = desc[f'{type}.set_rows']['value']
        ncol = desc[f'{type}.set_cols']['value']
        log.debug(f'rows: {nrow} x cols: {ncol} pixels')

        # Ground projected images
        # Delta latitude and longitude
        dlat = desc[f'{type}.row_mult']['value']
        dlon = desc[f'{type}.col_mult']['value']
        log.debug(f'latitude delta: {dlat}, longitude delta: {dlon} deg/pixel')
        # Upper left corner coordinates
        lat1 = desc[f'{type}.row_addr']['value']
        lon1 = desc[f'{type}.col_addr']['value']
        log.debug(f'Ref Latitude: {lat1}, Longitude: {lon1} degrees')

        # Lat1/lon1 are already the center so for geotiff were good to go.
        t = Affine.translation(float(lon1), float(lat1))* Affine.scale(float(dlon), float(dlat))

        # Get data type specific data
        bytes = desc[f'{type}.val_size']['value']
        endian = desc['val_endi']['value']
        log.debug(f'Bytes = {bytes}, Endian = {endian}')

        # Set up datatypes
        com_des = desc[f'{type}.val_frmt']['value']
        com = False
        if 'COMPLEX' in com_des:
            com = True
        log.debug(f'Complex descriptor {com_des}')
        if subtype == 'int' or type == 'int':
            dtype = np.dtype([('real', '<f4'), ('imaginary', '<f4')])
        else:
            if com:
                dtype = np.dtype([('real', '<f4'), ('imaginary', '<f4')])
            else:
                dtype = np.dtype([('real', '<f{}'.format(bytes))])
        log.debug(f'Data type = {dtype}')
        # Read in binary data
        z = np.fromfile(in_fp, dtype = dtype)

        # Reshape it to match what the text file says the image is
        z = z.reshape(nrow, ncol)

        # Build the transform and CRS
        crs = CRS.from_user_input("EPSG:4326")

        for comp in ['real', 'imaginary']:
            if comp in z.dtype.names:
                d = z[comp]
                # Change zeros and -10,000 to nans based on documentation.
                d[d==0]= np.nan
                d[d==-10000]= np.nan
                if type == 'slope':
                    d[d == np.nanmin(d)] = np.nan

                log.debug(f'Writing {comp} component to {out_fp}...')
                dataset = rasterio.open(
                    out_fp,
                    'w+',
                    driver='GTiff',
                    height=d.shape[0],
                    width=d.shape[1],
                    count=1,
                    dtype=d.dtype,
                    crs=crs,
                    transform=t,
                )
                # Write out the data
                dataset.write(d, 1)

                dataset.close()
        log.info('Finished converting image to WGS84 Geotiff.')

        return desc, z, type


"""
File control functions for uavsar_pytools.
"""

from zipfile import ZipFile
from tqdm import tqdm
import os
from os.path import exists, join

import logging
log = logging.getLogger(__name__)
logging.basicConfig()
log.setLevel(logging.DEBUG)

def unzip(dir_path, out_dir, pols = None):
    """
    Function to extract zipped directory with tqdm progress bar.
    From: https://stackoverflow.com/questions/4006970/monitor-zip-file-extraction-python.

    Args:
        dir_path (string) - path to zipped directory to unpack
        out_dir (string) - path to directory to extract files to.
    """
    assert exists(dir_path), f'Zipped directory at {dir_path} not found.'

    # Open your .zip file
    with ZipFile(file=dir_path) as zip_file:

        if pols:
            pol_list = [s for s in zip_file.namelist() if any(xs in s for xs in pols)]
            checked_list = [f for f in pol_list if not exists(join(out_dir, f))]
        else:
            pol_list = zip_file.namelist()
            checked_list = [f for f in pol_list if not exists(join(out_dir, f))]


        # Loop over each file
        if checked_list:
            for file in tqdm(iterable=checked_list, total=len(checked_list), unit = 'file', desc='Unzipping'):
                # Extract each file to another directory
                zip_file.extract(member=file, path=out_dir)
        else:
            log.info('No files found to unzip. Check if polarizations exist.')

    return [join(out_dir, fp) for fp in pol_list]


import os
import logging
import asf_search as asf
import pandas as pd

from uavsar_pytools.UavsarScene import UavsarScene

log = logging.getLogger(__name__)
logging.basicConfig()
log.setLevel(logging.DEBUG)

class UavsarCollection():
    """
    Class to handle uavsar collections containing many different image pairs. Methods include downloading and converting images.

    Args:
        collection (str): name of collection. Found at: https://api.daac.asf.alaska.edu/services/utils/mission_list
        work_dir (str): directory to download images into
        overwrite (bool): Do you want to overwrite pre-existing files [Default = False]
        clean (bool): Do you want to erase binary files after completion [Default = False]
        debug (str): level of logging (not yet implemented)
        pols (list): Do you want only certain polarizations? [Default = all available]
        dates (list): List of 1: start date and 2: end date to constrain collection results.
        low_ram (bool): decimates by a factor of 100 the arrays to conserve memory. [Default = True]

    Methods:
        collection_to_tiffs(): Main method. Finds all Uavsar Images in the collection and downloads, converts them to GeoTiffs.
    """

    Scenes = []

    def __init__(self, collection, work_dir, overwrite = False, clean = True, debug = False, pols = None, dates = None, low_ram = True):
        self.collection = collection
        self.work_dir = work_dir
        self.overwrite = overwrite
        self.clean = clean
        self.debug = debug
        self.pols = pols
        self.low_ram = low_ram
        if pols:
            pols = [pol.upper() for pol in pols]
            if set(pols).issubset(['VV','VH','HV','HH']):
                self.pols = pols
            else:
                raise ValueError('Bad Polarization Provided.')
        self.dates = dates
        if dates:
            # define search parameters for sierra flight line
            self.start_date = dates[0]
            self.end_date = dates[1]

    def find_urls(self):
        # search for data
        if self.dates:
            self.results = asf.search(platform = 'UAVSAR',
                        processingLevel = (['INTERFEROMETRY_GRD']),
                        campaign = self.collection,
                        start = self.start_date,
                        end = self.end_date)
        else:
            self.results = asf.search(platform = 'UAVSAR',
                        processingLevel = (['INTERFEROMETRY_GRD']),
                        campaign = self.collection)
        log.info(f'Found {len(self.results)} image pairs')

    def results_to_tiffs(self):
        for result in self.results:
            url = result.properties['url']
            log.info(f'Starting on: {url}')
            scene = UavsarScene(url = url, work_dir= self.work_dir, pols = self.pols, clean = self.clean, low_ram=self.low_ram)
            scene.url_to_tiffs()
            d1 = scene.images[0]['description']['start time of acquisition for pass 1']['value']
            d2 = scene.images[0]['description']['start time of acquisition for pass 2']['value']
            print(f'Completed {d1} to {d2}')

    def collection_to_tiffs(self):
        self.find_urls()
        self.results_to_tiffs()



"""
Originally written by HP Marshall in matlab. Transcribed by Micah J. into python. Amended for uavsar_pytools by Zach Keskinen.
Functions uses the urls to download uavsar data. It will not overwrite files
so if you want to re-download fresh manually remove the output_dir.
Warning: Canceling the script mid run will produce a file partially written. Rerunning the script will think the
file is already downloaded and skip it. You will have to remove that file if you want to re-download it.
"""

import requests
import os
from os.path import join, isdir, isfile, basename, dirname, exists
from tqdm.auto import tqdm
import logging

import time

log = logging.getLogger(__name__)
logging.basicConfig()
log.setLevel(logging.WARNING)

def stream_download(url, output_f):
    """
    Args:
        url: url to download
        output_f: path to save the data to
    """

    r = requests.get(url, stream=True)
    if r.status_code == 200:
        # Progress bar - https://towardsdatascience.com/how-to-download-files-using-python-part-2-19b95be4cdb5
        total_size= int(r.headers.get('content-length', 0))
        with open(output_f, 'wb') as f:
            with tqdm(total=total_size, unit='B', unit_scale=True , desc=f'Downloading {basename(url)}') as pbar:
                for ch in r.iter_content(chunk_size=1024):
                    if ch:
                        f.write(ch)
                        pbar.update(len(ch))
    else:
        if r.status_code == 401:
            log.warning(f'HTTP CODE 401. DOWNLOADING REQUIRES A NETRC FILE AND SIGNED UAVSAR END USER AGREEMENT!ß See ReadMe for instructions.')
        elif r.status_code == 404:
            log.warning(f'HTTP CODE 404. Url not found. Currently trying {url}.')
        else:
            log.warning(f'HTTP CODE {r.status_code}. Skipping download!')


def download_image(url, output_dir, ann = False, ann_url = None):
    """
    Downloads uavsar InSAR files from a url.
    Args:
        url (string): A url containing uavsar flight data. Can be from JPL or ASF
        output_dir (string): Directory to save the data in
    Returns:
        out_fp (string): File path to downloaded image.
    Raises:
       None
    """

    log.info(f'Starting download of {url}...')
    local = join(output_dir, basename(url))

    # Make the output dir if it doesn't exist
    if not isdir(output_dir):
        os.makedirs(output_dir)

    if not isfile(local):
        stream_download(url, local)
    else:
        log.info(f'{local} already exists, skipping download!')

    if ann:
        if ann_url == None:
            if url.split('.')[-1] == 'zip' or url.split('.')[-1] == 'ann':
                log.info('Download already contains ann file, skipping download!')
            else:
                # see if we can use the parent directory to extract the annotation file
                parent = dirname(url)
                # ASF formatting - query parent directory
                if parent.split('.')[-1] == 'zip':
                    log.debug(f'ASF url found for {url}')
                    parent_files = requests.get(parent).json()['response']
                    ann_info = [i for i in parent_files if '.ann' in i['name']][0]
                    # assert len(ann_info) == 1, 'More than one ann file detected'
                    ann_url = ann_info['url']
                    log.debug(f'Annotation url: {ann_url}')

                # Can't find zip parent directory - have to parse url to get ann
                else:
                    log.debug(f'Can not find zip parent directory.')
                    ext = url.split('.')[-1]
                    pols = ['VVVV','HHHH','HVHV', 'HHHV', 'HHVV','HVVV']
                    slc_pol = [pol for pol in pols if (pol in url)]
                    if len(slc_pol) == 1:
                        url = url.replace(slc_pol[0], '')

                    if ext == 'grd':
                        if len(basename(url).split('.')) == 2:
                            url = url.replace('.grd','.ann')
                        if len(basename(url).split('.')) == 3:
                            url = url.replace('.grd','')
                        ext = url.split('.')[-1]
                    elif ext == 'inc' and 'asf' in url:
                        url = url.replace('INC','METADATA')
                    ann_url = url.replace(f'.{ext}', '.ann')
                    log.debug(f'Parsed annotation url: {ann_url}')

                    response = requests.get(ann_url)
                    if response.status_code == 200:
                        log.debug('Success in parsing ann url')

                    else:
                        ann_url = None

                if ann_url:
                    ann_local = join(output_dir, basename(ann_url))
                    log.debug(f'Annotation local: {ann_local} and url {ann_url}')
                    if not isfile(ann_local):
                        stream_download(ann_url, ann_local)
                    else:
                        log.info(f'{ann_local} already exists, skipping download!')
                    return local, ann_local
                else:
                    log.warning('No ann url found. Manually provide .ann url.')
                    ann_url = None
        else:
            ann_local = join(output_dir, basename(ann_url))
            if not isfile(ann_local):
                stream_download(ann_url, ann_local)

        return local, None

def download_zip(url, output_dir):
    """
    Downloads uavsar InSAR files from a zip url.
    Args:
        url (string): A url containing uavsar flight zip. Can be from JPL or ASF
        output_dir (string): Directory to save the data in
    Returns:
        out_fp (string): File path to downloaded images.
    Raises:
       None
    """

    log.info(f'Starting download of {url}...')

    # Make the output dir if it doesn't exist
    if not isdir(output_dir):
        os.makedirs(output_dir)

    local = join(output_dir, basename(url))

    if not exists(local):
        stream_download(url, local)
    else:
        log.info(f'{local} already exists, skipping download!')

    return local
